/*
 * Copyright (c) 2000 Ping Pan and Columbia University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by Columbia University
 * and its contributors.'' Neither the name of the University nor the names 
 * of its contributors may be used to endorse or promote products derived 
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stddef.h>

#define _IP_VHL
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/proc.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/domain.h>
#include <sys/protosw.h>
#include <net/raw_cb.h>
#include <net/if.h>
#include <net/if_var.h>
#include <net/if_dl.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netipopt/ipopt_var.h>


static struct	sockaddr_in ipoptsrc = {sizeof(ipoptsrc), AF_IPOPTION};

void
ipopt_map(int opt, u_char *payload, struct ip *ip)
{
	bzero((char *)&ipopt_pkt, sizeof(struct ipopt_mask));

	switch (opt) {
	case IPOPT_RR:
		IPOPT_SET_RR(&ipopt_pkt);
		break;
	case IPOPT_TS:
		IPOPT_SET_TS(&ipopt_pkt);
		break;
	case IPOPT_SECURITY:
		IPOPT_SET_SEC(&ipopt_pkt);
		break;
	case IPOPT_LSRR:
		IPOPT_SET_LSRR(&ipopt_pkt);
		break;
	case IPOPT_SSRR:
		IPOPT_SET_SSRR(&ipopt_pkt);
		break;
	case IPOPT_RA: {
		struct  ip_routeralert *alert;
		u_short ra_val;

		alert = (struct ip_routeralert *)payload;
		ra_val = htons(alert->ipa_val);

		if (ra_val == IPOPT_RA_EXAM) {	/* RFC2113 */
			IPOPT_SET_RA(&ipopt_pkt);

			if (ip->ip_p == IPPROTO_RSVP)
				IPOPT_SET_ARSVP(&ipopt_pkt);
		}
		else if (ra_val == IPOPT_RA_RTCP) {
			IPOPT_SET_RA(&ipopt_pkt);
			IPOPT_SET_ARTCP(&ipopt_pkt);
		}

		break;
	}
	default:
		break;
	}
	return;
}


/* Return:
 *	-1	the packet's been processed and the buffer's been freed.
 *	0	didn't do anything; pass it back to the forwarding loop 
 */

int
ipopt_input(struct mbuf *m)
{
	struct ip *ip;
	struct mbuf *n;
	u_long opt_list, hlen, i, j;

	ip = mtod(m, struct ip *);
	hlen = IP_VHL_HL(ip->ip_vhl) << 2;

	opt_list = (ipopt_pkt.flag & ipopt_cfg.flag);
	if (opt_list == 0)		/* no match */
		return 0;

	/* is it the router alert pkt's that I want ? */
	if (opt_list & IPOPT_RA_MASK) {
		if (!(ipopt_pkt.alert_flag & ipopt_cfg.alert_flag))
			return 0;
	}

	for (i=1, j=0; j < IPOPT_MAX_OPTIONS; i <<=1,j++) {

		/* if match, give a copy to process */
		if (opt_list & i) {
			n = m_copy(m, 0, (int)M_COPYALL);
			if (n)
				(*ipoptsw[ipopt_protox[j]].pr_input)(n, hlen);
			else
				ipoptstat.no_space++;
		}
	}
	m_freem(m);
	return -1;
}


void
null_input(struct mbuf *m, int iphlen)
{
	ipoptstat.null_input++;
	m_freem(m);
	return;
}

void
rr_input(struct mbuf *m, int iphlen)
{
	ipoptstat.rr_input++;
	ipopt_proto_input(m, iphlen, IPOPT_RR);
}

void
ts_input(struct mbuf *m, int iphlen)
{
	ipoptstat.ts_input++;
	ipopt_proto_input(m, iphlen, IPOPT_TS);
}

void
security_input(struct mbuf *m, int iphlen)
{
	ipoptstat.sec_input++;
	ipopt_proto_input(m, iphlen, IPOPT_SECURITY);
}

void
lsrr_input(struct mbuf *m, int iphlen)
{
	ipoptstat.lsrr_input++;
	ipopt_proto_input(m, iphlen, IPOPT_LSRR);
}

void
ssrr_input(struct mbuf *m, int iphlen)
{
	ipoptstat.ssrr_input++;
	ipopt_proto_input(m, iphlen, IPOPT_SSRR);
}


void
ra_input(struct mbuf *m, int iphlen)
{
	register struct ip *ip = mtod(m, struct ip *);
	register struct ipoptpcb *ocb, *last;
	struct mbuf *opts;

	opts = 0;
	last = 0;
	ipoptstat.ra_input++;
	ipoptsrc.sin_addr = ip->ip_src;

	LIST_FOREACH(ocb, &ipoptcb_list, list) {
		if (ocb->ocb_proto.sp_protocol != IPOPT_RA)
			continue;

		if (IPOPT_ARSVP_SET(&ipopt_pkt) && (!OCB_RSVP_SET(ocb)))
			continue;

		if (IPOPT_ARTCP_SET(&ipopt_pkt) && (!OCB_RTCP_SET(ocb)))
			continue;

		if (last) {
			struct mbuf *n;
			n = m_copy(m, 0, (int)M_COPYALL);
			if (n) {
				if (last->ocb_flags)
					ipopt_appenctl(last, &opts, ip, n);

				if (sbappendaddr(&last->ocb_socket->so_rcv, 
				    (struct sockaddr *)&ipoptsrc, n,
				    opts) == 0)  {
					ipoptstat.lost_input++;
					m_freem(n);
					if (opts) 
						m_freem(opts);
				}
				else {
					sorwakeup(last->ocb_socket);
				}
			}
		}
		last = ocb;
	}
	if (last) {
		if (last->ocb_flags)
			ipopt_appenctl(last, &opts, ip, m);

		if (sbappendaddr(&last->ocb_socket->so_rcv, 
		    (struct sockaddr *)&ipoptsrc, m, opts) == 0) {
			ipoptstat.lost_input++;
			m_freem(m);
			if (opts)
				m_freem(opts);
		}
		else {
			sorwakeup(last->ocb_socket);
		}
	} else {
		m_freem(m);
	}
}



/* 
 * generic IP Option input
 */
void
ipopt_proto_input(struct mbuf *m, int iphlen, int proto)
{
	register struct ip *ip = mtod(m, struct ip *);
	register struct ipoptpcb *ocb, *last;
	struct mbuf *opts;

	opts = 0;
	last = 0;
	ipoptsrc.sin_addr = ip->ip_src;

	LIST_FOREACH(ocb, &ipoptcb_list, list) {
		if (ocb->ocb_proto.sp_protocol != proto)
			continue;

		if (last) {
			struct mbuf *n;
			n = m_copy(m, 0, (int)M_COPYALL);
			if (n) {
				if (last->ocb_flags)
					ipopt_appenctl(last, &opts, ip, n);

				if (sbappendaddr(&last->ocb_socket->so_rcv, 
				    (struct sockaddr *)&ipoptsrc, n,
				    opts) == 0)  {
					ipoptstat.lost_input++;
					m_freem(n);
					if (opts) 
						m_freem(opts);
				}
				else {
					sorwakeup(last->ocb_socket);
				}
			}
		}
		last = ocb;
	}
	if (last) {
		if (last->ocb_flags)
			ipopt_appenctl(last, &opts, ip, m);

		if (sbappendaddr(&last->ocb_socket->so_rcv, 
		    (struct sockaddr *)&ipoptsrc, m, opts) == 0) {
			ipoptstat.lost_input++;
			m_freem(m);
			if (opts)
				m_freem(opts);
		}
		else {
			sorwakeup(last->ocb_socket);
		}
	} else {
		m_freem(m);
	}
}


/* copied from ip_input.c... don't hate me 'cause the goto */
void
ipopt_appenctl(ocb, mp, ip, m)
	register struct ipoptpcb *ocb; 
	register struct mbuf **mp;
	register struct ip *ip;
	register struct mbuf *m;
{
	struct ifnet *ifp;
	struct sdlbuf {
		struct sockaddr_dl sdl;
		u_char	pad[32];
	} sdlbuf;
	struct sockaddr_dl *sdp;
	struct sockaddr_dl *sdl2 = &sdlbuf.sdl;
	int n;

	if (!OCB_IF_SET(ocb))		/* only support IP_RECVIF now */
		return;

	ifp = (struct ifnet *)(m->m_pkthdr.rcvif);
	if (ifp && ifp->if_index && (ifp->if_index <= if_index)) {

		n = ifp->if_index;
		sdp = (struct sockaddr_dl *)(ifnet_addrs[n-1]->ifa_addr);

		/* Change our mind and don't try copy. */
		if ((sdp->sdl_family != AF_LINK)
		|| (sdp->sdl_len > sizeof(sdlbuf))) {
			goto makedummy;
		}
		bcopy(sdp, sdl2, sdp->sdl_len);
	}
	else {
makedummy:	
		sdl2->sdl_len = offsetof(struct sockaddr_dl, sdl_data[0]);
		sdl2->sdl_family = AF_LINK;
		sdl2->sdl_index = 0;
		sdl2->sdl_nlen = sdl2->sdl_alen = sdl2->sdl_slen = 0;
	}

	*mp = sbcreatecontrol((caddr_t) sdl2, sdl2->sdl_len,
		IP_RECVIF, IPPROTO_IP);
	if (*mp)
		mp = &(*mp)->m_next;
}

