/*
 * Copyright (c) 2000 Ping Pan and Columbia University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by Columbia University
 * and its contributors.'' Neither the name of the University nor the names 
 * of its contributors may be used to endorse or promote products derived 
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/proc.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/domain.h>
#include <sys/protosw.h>
#include <net/raw_cb.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netipopt/ipopt_var.h>

/* all ipopt extension variables */
struct ipopt_stat ipoptstat;
struct ipopt_mask ipopt_cfg;
struct ipopt_mask ipopt_pkt;
u_char ipopt_protox[IPOPT_MAX_OPTIONS];
struct ipoptcb_list_head ipoptcb_list;
static u_long	ipopt_sendspace = IPOPTSNDQ;
static u_long	ipopt_recvspace = IPOPTRCVQ;


/* functions in protosw and doamin */

void
ipopt_init()
{
	bzero((char *)&ipoptstat, sizeof(struct ipopt_stat));
	bzero((char *)&ipopt_cfg, sizeof(struct ipopt_mask));
	bzero((char *)&ipopt_protox, IPOPT_MAX_OPTIONS * sizeof(u_char));

	ipopt_reg_proto(IPOPT_RR, IPOPT_RR_OFFSET);
	ipopt_reg_proto(IPOPT_TS, IPOPT_TS_OFFSET);
	ipopt_reg_proto(IPOPT_SECURITY, IPOPT_SEC_OFFSET);
	ipopt_reg_proto(IPOPT_LSRR, IPOPT_LSRR_OFFSET);
	ipopt_reg_proto(IPOPT_SSRR, IPOPT_SSRR_OFFSET);
	ipopt_reg_proto(IPOPT_RA, IPOPT_RA_OFFSET);

	LIST_INIT(&ipoptcb_list);
}


/* ipopt socket option processing
 */
int
ipopt_ctloutput(struct socket *so, struct sockopt *sopt)
{
	struct ipoptpcb *ocb = sotoipoptcb(so);
	int error, val;

	if (ocb == 0)
		return EINVAL;
	if (sopt->sopt_level != IPPROTO_IP)
		return EINVAL;

	error = val = 0;

	switch (sopt->sopt_dir) {
	case SOPT_SET:
		error = sooptcopyin(sopt, &val, sizeof val, sizeof val);
		if (error)
			break;

		switch (sopt->sopt_name) {
		case IP_RECVIF:
			OCB_SET_RECVIF(ocb);
			break;

		case IPOPT_RECVRTCP:
			if (ocb->ocb_proto.sp_protocol != IPOPT_RA) {
				error = EINVAL;
				break;
			}
			ipoptstat.rtcp_count++;
			IPOPT_SET_ARTCP(&ipopt_cfg);
			OCB_SET_RTCP(ocb);
			break;

		case IPOPT_RECVRSVP:
			if (ocb->ocb_proto.sp_protocol != IPOPT_RA) {
				error = EINVAL;
				break;
			}
			ipoptstat.rsvp_count++;
			IPOPT_SET_ARSVP(&ipopt_cfg);
			OCB_SET_RSVP(ocb);
			break;

		default:
			error = ENOPROTOOPT;
			break;
		}
		break;
	
	case SOPT_GET:
		switch (sopt->sopt_name) {
		case IP_RECVIF:
			val = OCB_BIT(ocb, OCB_RECVIF);
			break;

		case IPOPT_RECVRTCP:
			if (ocb->ocb_proto.sp_protocol != IPOPT_RA) {
				error = EINVAL;
				break;
			}
			ipoptstat.rsvp_count++;
			val = OCB_BIT(ocb, OCB_RECVRTCP);
			break;

		case IPOPT_RECVRSVP:
			if (ocb->ocb_proto.sp_protocol != IPOPT_RA) {
				error = EINVAL;
				break;
			}
			ipoptstat.rsvp_count++;
			val = OCB_BIT(ocb, OCB_RECVRSVP);
			break;

		default:
			error = ENOPROTOOPT;
			break;
		}

		if (!error)
			error = sooptcopyout(sopt, &val, sizeof val); 
		break;
	}
	return error;
}


/* functions in pr_usrreqs */

/* disconnect the control block, release and disconnect the socket;
 */
static int
ipopt_abort(struct socket *so)
{
	struct ipoptpcb *ocb = sotoipoptcb(so);
	int s, error;

	if (ocb == 0)
		return EINVAL;

	s = splnet();
	ipopt_reset_cb(so);
	error = ipopt_free_cb(so);
	soisdisconnected(so);
	splx(s);

	return error;
}


/* allocate and insert a control block, and socket buffer space
 */

static int
ipopt_attach(struct socket *so, int proto, struct proc *p)
{
	struct ipoptpcb *ocb;
	int s, error;

	if (sotoipoptcb(so) != 0)
		return EISCONN;

	if (p && (error = suser(p)) != 0)
		return error;

	MALLOC(ocb, struct ipoptpcb *, sizeof *ocb, M_PCB, M_WAITOK);
	if (ocb == 0)
		return ENOBUFS;
	bzero(ocb, sizeof *ocb);

	s = splnet();
	so->so_pcb = (caddr_t)ocb;
	error = soreserve(so, ipopt_sendspace, ipopt_recvspace);
	if (error) {
		splx(s);
		free(ocb, M_PCB);
		return (error);
	}

	ocb->ocb_socket = so;
	ocb->ocb_proto.sp_family = so->so_proto->pr_domain->dom_family;
	ocb->ocb_proto.sp_protocol = proto;
	LIST_INSERT_HEAD(&ipoptcb_list, ocb, list);

	error = ipopt_accept_proto(proto);
	if (error) {
		splx(s);
		free(ocb, M_PCB);
		return error;
	}
	soisconnected(so);
	splx(s);

	return 0;
}


/* do nothing for raw */
static int
ipopt_bind(struct socket *so, struct sockaddr *nam, struct proc *p)
{
	return EINVAL;
}

/* do nothing for raw */
static int
ipopt_connect(struct socket *so, struct sockaddr *nam, struct proc *p)
{
	return EINVAL;
}


/* release the socket... just like ipopt_abort() */
static int
ipopt_detach(struct socket *so)
{
	struct ipoptpcb *ocb = sotoipoptcb(so);
	int s, error;

	if (ocb == 0)
		return EINVAL;

	s = splnet();
	ipopt_reset_cb(so);
	error = ipopt_free_cb(so);
	splx(s);

	return error;
}

/* same as detach */
static int
ipopt_disconnect(struct socket *so)
{
	struct ipoptpcb *ocb = sotoipoptcb(so);
	int s, error;

	if (ocb == 0)
		return EINVAL;

	s = splnet();
	ipopt_reset_cb(so);
	error = ipopt_free_cb(so);
	splx(s);

	return error;
}


/* for now, I don't support remote address setting 
 */
static int
ipopt_peeraddr(struct socket *so, struct sockaddr **nam)
{
	return EINVAL;
}


/* process writing to the socket */
static int
ipopt_send(struct socket *so, int flags, struct mbuf *m, struct sockaddr *nam,
	 struct mbuf *control, struct proc *p)
{
	int s;

	s = splnet();

	if (m != NULL)
		m_freem(m);

	splx(s);

	return EINVAL;
}

static int
ipopt_shutdown(struct socket *so)
{
	struct ipoptpcb *ocb = sotoipoptcb(so);
	int s;

	if (ocb == 0)
		return EINVAL;

	s = splnet();
	socantsendmore(so);
	splx(s);

	return 0;
}

/* make getsockname call working */
static int
ipopt_sockaddr(struct socket *so, struct sockaddr **nam)
{
	return EINVAL;
}


struct pr_usrreqs	ipopt_usrreqs = {
	ipopt_abort, pru_accept_notsupp, ipopt_attach, ipopt_bind, 
	ipopt_connect, pru_connect2_notsupp, pru_control_notsupp, 
	ipopt_detach, ipopt_disconnect, pru_listen_notsupp, ipopt_peeraddr, 
	pru_rcvd_notsupp, pru_rcvoob_notsupp, ipopt_send, pru_sense_null, 
	ipopt_shutdown, ipopt_sockaddr,
	sosend, soreceive, sopoll
};


/* utility functions */
/* register protocol func's to a map for quick data forwarding 
 */
void
ipopt_reg_proto(int proto, int offset)
{ 
	struct protosw *pr;

	pr = pffindproto(PF_IPOPTION, proto, SOCK_RAW);
	if (pr == 0)
		ipopt_protox[offset] = 0;	/* default one */
	else
		ipopt_protox[offset] = pr - ipoptsw;
}


/* reset the system parameter 
 */
void
ipopt_reset_cb(struct socket *so)
{
	struct ipoptpcb *ocb = sotoipoptcb(so);

	switch(ocb->ocb_proto.sp_protocol) {
	case IPOPT_RR:
		ipoptstat.rr_count--;
		if (!ipoptstat.rr_count)
			IPOPT_RESET_RR(&ipopt_cfg);
		break;
	case IPOPT_TS:
		ipoptstat.ts_count--;
		if (!ipoptstat.ts_count)
			IPOPT_RESET_TS(&ipopt_cfg);
		break;
	case IPOPT_SECURITY:
		ipoptstat.sec_count--;
		if (!ipoptstat.sec_count)
			IPOPT_RESET_SEC(&ipopt_cfg);
		break;
	case IPOPT_LSRR:
		ipoptstat.lsrr_count--;
		if (!ipoptstat.lsrr_count)
			IPOPT_RESET_LSRR(&ipopt_cfg);
		break;
	case IPOPT_SSRR:
		ipoptstat.ssrr_count--;
		if (!ipoptstat.ssrr_count)
			IPOPT_RESET_SSRR(&ipopt_cfg);
		break;
	case IPOPT_RA:
		ipoptstat.ra_count--;
		if (!ipoptstat.ra_count)
			IPOPT_RESET_RA(&ipopt_cfg);

		if (OCB_RTCP_SET(ocb)) {
			ipoptstat.rtcp_count--;
			if (!ipoptstat.rtcp_count) {
				IPOPT_RESET_ARTCP(&ipopt_cfg);
			}
		}

		if (OCB_RSVP_SET(ocb)) {
			ipoptstat.rsvp_count--;
			if (!ipoptstat.rsvp_count) {
				IPOPT_RESET_ARSVP(&ipopt_cfg);
			}
		}
		break;
	}
}


/* remove the control block from the chain list and free the memory
 */
int
ipopt_free_cb(struct socket *so)
{
	struct ipoptpcb *ocb = sotoipoptcb(so);

	if (ocb == 0)
		return EINVAL;
	if (ocb->ocb_socket != so)	/* huh? */
		return EINVAL;

	so->so_pcb = 0;
	sofree(so);
	LIST_REMOVE(ocb, list);
	free((caddr_t)(ocb), M_PCB);
	return 0;
}


int
ipopt_accept_proto(int proto)
{
	int error=0;

	switch(proto) {
	case IPOPT_RR:
		ipoptstat.rr_count++;
		IPOPT_SET_RR(&ipopt_cfg);
		break;

	case IPOPT_TS:
		ipoptstat.ts_count++;
		IPOPT_SET_TS(&ipopt_cfg);
		break;

	case IPOPT_SECURITY:
		ipoptstat.sec_count++;
		IPOPT_SET_SEC(&ipopt_cfg);
		break;

	case IPOPT_LSRR:
		ipoptstat.lsrr_count++;
		IPOPT_SET_LSRR(&ipopt_cfg);
		break;

	case IPOPT_SSRR:
		ipoptstat.ssrr_count++;
		IPOPT_SET_SSRR(&ipopt_cfg);
		break;

	case IPOPT_RA:
		ipoptstat.ra_count++;
		IPOPT_SET_RA(&ipopt_cfg);
		break;

	default:
		error = EPROTONOSUPPORT;
	}
	return error;
}
